"""Buffer representation and writer's logic"""
import functools
import logging
import operator
import traceback
from decimal import Decimal
from mmap import mmap
from struct import pack_into, error as struct_error, pack
from typing import Any, List, Union, Tuple

import numpy as np

from .casting import date_to_int, datetime_to_long, decimal_to_sq_numeric
from .globals import WIN, ROWS_PER_FLUSH, VARCHAR_ENCODING, \
    type_to_letter, ARROW, buf_maps, PYTYPES, typecodes
from .utils import DataError, ProgrammingError
from .logger import log_and_raise, logger, printdbg


class ColumnBuffer:
    ''' Buffer holding packed columns to be sent to SQream '''

    def __init__(self, *_, **__):
        ...

    def clear(self):
        if buf_maps:
            [buf_map.close() for buf_map in buf_maps[0]]

    def pack_columns(self, cols, capacity, col_types, col_sizes, col_nul, col_tvc, col_scales):
        ''' Packs the buffer starting a given index with the column.
            Returns number of bytes packed '''

        pool_params = list(zip(cols, range(len(col_types)), col_types,
                          col_sizes, col_nul, col_tvc, col_scales))

        if WIN:
            packed_cols = []
            for param_tup in pool_params:
                packed_cols.append(_pack_column(param_tup))

        else:
            try:
                packed_cols = [_pack_column(x) for x in pool_params]
            except DataError:
                raise  # Expected error, shouldn't be caught and wrapped
            except Exception as e:
                printdbg("Original error from pool.map: ", e)
                if logger.isEnabledFor(logging.ERROR):
                    logger.error("Original error from pool.map: ", e)
                log_and_raise(ProgrammingError,
                              "Error packing columns. Check that all types match the respective column types")

        return list(packed_cols)

    def close(self):
        self.clear()


## A top level packing function for Python's MP compatibility
def _pack_column(col_tup, return_actual_data=True):
    ''' Packs the buffer starting a given index with the column.
        Returns number of bytes packed '''

    global CYTHON
    col, col_idx, col_type, size, nullable, tvc, scale = col_tup
    if "ftArray" in col_type:
        return _pack_array(col, col_idx, col_type[1], nullable, scale)

    col = list(col)
    capacity = len(col)
    buf_idx = 0
    # Don't need to initialize bit mmap, could be created after gathering
    # dynamic array (list) directly into bytes
    # TODO: replace logic using list
    buf_map = mmap(-1, ((1 if nullable else 0) + (size if size != 0 else 104)) * ROWS_PER_FLUSH)
    buf_view = memoryview(buf_map)

    def pack_exception(e):
        ''' Allowing to return traceback info from parent process when using mp.Pool on _pack_column
            [add link]
        '''

        e.traceback = traceback.format_exc()
        error_msg = f'Trying to insert unsuitable types to column number {col_idx + 1} of type {col_type}'
        logger.error(error_msg, exc_info=True)
        raise ProgrammingError(error_msg)

    # Numpy array for column
    if ARROW and isinstance(col, np.ndarray):
        # Pack null column if applicable
        if nullable:
            pack_into(f'{capacity}b', buf_view, buf_idx,
                      *[1 if item in (np.nan, b'') else 0 for item in col])
            buf_idx += capacity

            # Replace Nones with appropriate placeholder
            if 'S' in repr(col.dtype):  # already b''?
                pass
            elif 'U' in repr(col.dtype):
                pass
            else:
                # Swap out the nans
                col[col == np.nan] = 0

        # Pack nvarchar length column if applicable
        if tvc:
            buf_map.seek(buf_idx)
            lengths_as_bytes = np.vectorize(len)(col).astype('int32').tobytes()
            buf_map.write(lengths_as_bytes)
            buf_idx += len(lengths_as_bytes)

        # Pack the actual data
        if 'U' in repr(col.dtype):
            packed_strings = ''.join(col).encode('utf8')
            buf_map.seek(buf_idx)
            buf_map.write(packed_strings)
            buf_idx += len(packed_strings)
            # print(f'unicode strings: {packed_strings}')
        else:
            packed_np = col.tobytes()
            buf_map.seek(buf_idx)
            buf_map.write(packed_np)
            buf_idx += len(packed_np)

        return buf_map[0:buf_idx] if return_actual_data else (0, buf_idx)

    # Pack null column if applicable
    type_code = type_to_letter[col_type]

    # Pack null column and replace None with appropriate placeholder
    col_placeholder = {
        'ftBool': 0,
        'ftUByte': 0,
        'ftShort': 0,
        'ftInt': 0,
        'ftLong': 0,
        'ftFloat': 0,
        'ftDouble': 0,
        'ftDate': None,  # updated separately
        'ftDateTime': None,  # updated separately
        'ftVarchar': ''.ljust(size, ' '),
        'ftBlob': '',
        'ftNumeric': 0
    }

    if nullable:
        idx = -1
        # TODO: Replace while True and index with just traversal without idx
        # for better time complexity
        while True:
            try:
                idx = col.index(None, idx + 1)
            except ValueError:
                break
            else:
                buf_map.seek(buf_idx + idx)
                buf_map.write(b'\x01')
                col[idx] = col_placeholder[col_type]
        # buf_map.seek(buf_idx)
        # buf_map.write(nulls)
        buf_idx += capacity

    # If a text column, replace and pack in adavnce to see if the buffer is sufficient
    if col_type == 'ftBlob':
        try:
            encoded_col = [strn.encode('utf8') for strn in col]
        except AttributeError as e:  # Non strings will not have .encode()
            pack_exception(e)
        else:
            packed_strings = b''.join(encoded_col)

        needed_buf_size = len(packed_strings) + 5 * capacity

        # Resize the buffer if not enough space for the current strings
        if needed_buf_size > len(buf_map):
            buf_view.release()
            buf_map.resize(needed_buf_size)
            buf_view = memoryview(buf_map)

        # Pack nvarchar length column
        pack_into(f'{capacity}i', buf_view, buf_idx, *[len(string) for string in encoded_col])

        buf_idx += 4 * capacity

    # Replace Nones with appropriate placeholder - this affects the data itself
    if col_type == 'ftBlob':
        pass  # Handled preemptively due to allow possible buffer resizing

    elif col_type == 'ftVarchar':
        try:
            col = (strn.encode(VARCHAR_ENCODING)[:size].ljust(size, b' ') for strn in col)
        except AttributeError as e:  # Non strings will not have .encode()
            pack_exception(e)
        else:
            packed_strings = b''.join(col)

    elif col_type == 'ftDate':
        try:
            col = (date_to_int(deit) for deit in col)
        except AttributeError as e:  # Non date/times will not have .timetuple()
            pack_exception(e)

    elif col_type == 'ftDateTime':
        try:
            col = (datetime_to_long(dt) for dt in col)
        except AttributeError as e:
            pack_exception(e)

    elif col_type == 'ftNumeric':
        try:
            col = (decimal_to_sq_numeric(Decimal(num), scale) for num in col)
        except AttributeError as e:
            pack_exception(e)

    elif col_type in ('ftBool', 'ftUByte', 'ftShort', 'ftInt', 'ftLong', 'ftFloat', 'ftDouble'):
        pass
    else:
        error_msg = f'Bad column type passed: {col_type}'
        logger.error(error_msg, exc_info=True)
        raise ProgrammingError(error_msg)

    CYTHON = False
    # Done preceding column handling, pack the actual data
    if col_type in ('ftVarchar', 'ftBlob'):
        buf_map.seek(buf_idx)
        buf_map.write(packed_strings)
        buf_idx += len(packed_strings)
    elif col_type == 'ftNumeric':
        buf_map.seek(buf_idx)
        all = functools.reduce(operator.iconcat, (num.to_bytes(16, byteorder='little', signed=True) for num in col), [])
        buf_map.write(bytearray(all))
        buf_idx += len(all)
    else:
        try:
            if CYTHON:
                buf_map.seek(buf_idx)
                # type_packer[col_type](col, size, buf_map, buf_idx)
            else:
                pack_into(f'{capacity}{type_code}', buf_view, buf_idx, *col)
        except struct_error as e:
            pack_exception(e)

        buf_idx += capacity * size

    return buf_map[0:buf_idx] if return_actual_data else (0, buf_idx)


def _pack_array(col: List[Union[List[Any], None]],
                col_idx: int,
                col_type: str,
                nullable: bool,
                scale: int
                ):
    """Pack array of any data type to bytes for transmitting to SQREAM

    Acceptable format of bytes includes 3 parts connected together
    without any padding gaps between:
        part1: nulls if nullable just N elements without padding
        part2: lengths of data chunks in part3 for each N row
            just N integers without padding (mem N * 4)
        part3: data without padding (varies depend on type of array)

    Explanation on part3 for fixed SIZE data types:
        First come 1 byte that explains whether data at first index is
        a null or not, then come SIZE bytes with the value, then again
        1 byte and SIZE bytes etc. When row ends, then just new array
        started without padding by the same pattern.

        Example of ARRAY INT[] of 3 rows:
        array[1, null, 8], null, array[2, 5]
        Packed: `010  15000 0000 10000  0 1000 1 0000 0 8000  0 2000 0 5000`
                 ^    ^                 ^                     ^
            p1 nulls  p2 lengths        p3 data row1          row3

    Explanation on part3 for unfixed size data type (TEXT):
        After part1 & 2 (the same)
        Started with array size in first 4 bytes (INT value)
        Then for each N elements: 1 byte (is null), 4 byte - INT represents
        LENGTH of string at this index, then LENGTH bytes data of string
        itself. Then the same for each row

        Example of ARRAY TEXT[] of 3 rows:
        array['WORK', '', null, "IS DONE"], null, array['Ok', 'then']
        Full buffer:
            part1 and 2: `010  35 000 0000 20 000`
                          ^    ^
                     p1 nulls  p2 lengths
            row1: `4000  0 4000 WORK  0 0000  1 0000  0 7000 IS DONE`
                   ^     ^            ^       ^       ^
            Array size   idx=1        idx=2   idx=3   idx=4
            row2: ``  - empty because it is null
            row3: `2000  0 2000 Ok  0 4000 then`
                   ^     ^          ^
            Array size   idx=1      idx=2

    Args:
        col: list of rows that could be lists that represents array or None
        col_idx: integer index of inserting column,
        col_type: string (enum) that represents data type that array contains
          could be ftBool, ftInt, etc
        nullable: boolean that show whether column is nullable or not,
        scale: integer represent Numerical Rating Scale, for numerical only

    Returns:
        A bytes of packed array that could be sent to SQREAM

        Examples from description:
        1. b'\x00\x01\x00\x0f\x00\x00\x00\x00\x00\x00\x00\n\x00\x00\x00'
           b'\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x08\x00\x00'
           b'\x00\x00\x02\x00\x00\x00\x00\x05\x00\x00\x00'
        2. b'\x00\x01\x00#\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00'
           b'\x04\x00\x00\x00\x00\x04\x00\x00\x00WORK\x00\x00\x00\x00'
           b'\x00\x01\x00\x00\x00\x00\x00\x07\x00\x00\x00IS DONE\x02'
           b'\x00\x00\x00\x00\x02\x00\x00\x00Ok\x00\x04\x00\x00\x00then'

    Raises:
      DataError: If None row is present for not nullable column
      DataError: If type value inside array (list) does not match
        appropriate type for column and it is not None
    """
    py_types = PYTYPES[col_type]

    trasform: callable = _get_backward_transform_func(col_type, scale)

    nulls = []
    lengths = []
    data_list = []
    for row in col:
        if nullable:
            nulls.append(pack('?', row is None))
        elif row is None:
            raise DataError(
                f"There is null in data for not nullable column {col_idx}")

        d_size = 0

        if row is not None and len(row) > 0:
            if typecodes[col_type] == 'STRING':
                data_list.append(pack('i', len(row)))
                d_size = 4
            for val in row:
                is_null = val is None
                if not is_null and not isinstance(val, py_types):
                    raise DataError(
                        f"Array contains inappropriate type: {type(val)}. "
                        f"Only None and {py_types} are allowed")
                data_list.append(pack('?', is_null))
                data_list.append(trasform(val))
                # Counting data size in runtime allows to use the same
                # code of both fixed and unfixed data types
                # len is always O(1)
                d_size += len(data_list[-1]) + 1
        lengths.append(pack('i', d_size))

    res = b''.join(nulls + lengths + data_list)
    return res


def _get_backward_transform_func(col_type: Union[Tuple[str], str], scale: int):
    """
    Provide function for casting py data to bytes that SQREAM expects

    Args:
        col_type: string (enum) that represent SQREAM data type
          OR tuple with strings that represent ftArray and its' data type
        scale: integer represent Numerical Rating Scale, for numerical only

    Returns:
        A callable that take one value - data and transform it into
          SQREAM compatible bytes
    """

    # Array's type_tup differs from others by adding extra string
    # at the beginning
    if 'ftArray' in col_type:
        return _get_backward_transform_func(col_type[1], scale)

    data_format = type_to_letter[col_type]
    wrappers = {
        'ftDate': date_to_int,
        'ftDateTime': datetime_to_long,
    }

    if col_type == 'ftNumeric':
        def cast(data: Any) -> bytes:
            val = decimal_to_sq_numeric(data, scale) if data else 0
            return val.to_bytes(16, byteorder='little', signed=True)
    elif col_type in wrappers:
        def cast(data: Any) -> bytes:
            return pack(data_format, wrappers[col_type](data))
    elif typecodes[col_type] == 'STRING':
        def cast(data: str) -> bytes:
            if not data:
                return b'\x00\x00\x00\x00'
            return pack(f'i{len(data)}s', len(data), data.encode())
    else:
        def cast(data: Any) -> bytes:
            return pack(data_format, data or 0)

    return cast



"""
Support functions for converting py values to sqream compatible values
and vice versa
"""
from datetime import datetime, date
from decimal import Decimal, getcontext
from math import floor, ceil, pow


def pad_dates(num):
    return ('0' if num < 10 else '') + str(num)


def sq_date_to_py_date(sqream_date, is_null=False, date_convert_func=date):

    if is_null:
        return None

    year = (10000 * sqream_date + 14780) // 3652425
    intermed_1 = 365 * year + year // 4 - year // 100 + year // 400
    intermed_2 = sqream_date - intermed_1
    if intermed_2 < 0:
        year = year - 1
        intermed_2 = sqream_date - (365 * year + year // 4 - year // 100 +
                                    year // 400)
    intermed_3 = (100 * intermed_2 + 52) // 3060

    year = year + (intermed_3 + 2) // 12
    month = int((intermed_3 + 2) % 12) + 1
    day = int(intermed_2 - (intermed_3 * 306 + 5) // 10 + 1)

    return date_convert_func(year, month, day)


def sq_datetime_to_py_datetime(sqream_datetime, is_null=False, dt_convert_func=datetime):
    ''' Getting the datetime items involves breaking the long into the date int and time it holds
        The date is extracted in the above, while the time is extracted here  '''

    if is_null:
        return None

    date_part = sqream_datetime >> 32
    time_part = sqream_datetime & 0xffffffff
    date_part = sq_date_to_py_date(date_part, is_null=is_null)

    if date_part is None:
        return None

    msec = time_part % 1000
    sec = (time_part // 1000) % 60
    mins = (time_part // 1000 // 60) % 60
    hour = time_part // 1000 // 60 // 60
    return dt_convert_func(date_part.year, date_part.month, date_part.day,
                           hour, mins, sec, msec * int(pow(10, 3)))  # Python expects to get 6 digit on
                                                                     # miliseconds while SQream returns 3.


def _get_date_int(year: int, month: int, day: int) -> int:
    """Convert year, month and day to integer compatible with SQREAM"""
    month: int = (month + 9) % 12
    year: int = year - month // 10
    return (
        365 * year + year // 4 - year // 100 + year // 400
        + (month * 306 + 5) // 10 + (day - 1)
    )


def date_to_int(dat: date) -> int:
    """Convert datetime.date to integer compatible with SQREAM interface"""
    # datetime is also supported because it is descendant of date
    # date_to_int(date(1900, 1, 1)) is 693901 which is the oldest date that
    # sqream supports, so for None use the same
    return 693901 if dat is None else _get_date_int(*dat.timetuple()[:3])


def datetime_to_long(dat: datetime) -> int:
    """Convert datetime.datetime to integer (LONG) compatible with SQREAM"""
    if dat is None:
        # datetime_to_long(datetime(1900, 1, 1)) is 2980282101661696 which is
        # the oldest date that sqream supports, so for None use the same
        return 2980282101661696
    year, month, day, hour, minute, second = dat.timetuple()[:6]
    msec = dat.microsecond

    date_int: int = _get_date_int(year, month, day)
    time_int: int = 1000 * (hour * 3600 + minute * 60 + second) + msec // 1000

    return (date_int << 32) + time_int


tenth = Decimal("0.1")
if getcontext().prec < 38:
    getcontext().prec = 38


def sq_numeric_to_decimal(bigint_as_bytes: bytes, scale: int, is_null=False) -> [Decimal, None]:
    if is_null:
        return None

    getcontext().prec = 38
    c = memoryview(bigint_as_bytes).cast('i')
    bigint = ((c[3] << 96) + ((c[2] & 0xffffffff) << 64) + ((c[1] & 0xffffffff) << 32) + (c[0] & 0xffffffff))

    return Decimal(bigint) * (tenth ** scale)


def decimal_to_sq_numeric(dec: Decimal, scale: int) -> int:  # returns bigint
    if getcontext().prec < 38:
        getcontext().prec = 38
    res = dec * (10 ** scale)
    return ceil(res) if res > 0 else floor(res)


def lengths_to_pairs(nvarc_lengths):
    ''' Accumulative sum generator, used for parsing nvarchar columns '''

    idx = new_idx = 0
    for length in nvarc_lengths:
        new_idx += length
        yield idx, new_idx
        idx = new_idx


def arr_lengths_to_pairs(text_lengths):
    """Generator for parsing ARRAY TEXT columns' data"""
    start = 0
    for length in text_lengths:
        yield start, length
        start = length + (8 - length % 8) % 8


def numpy_datetime_str_to_tup(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    numpy_dt = repr(numpy_dt).split("'")[1]
    date_part, time_part = numpy_dt.split('T')
    year, month, day = date_part.split('-')
    hms, ns = time_part.split('.')
    hour, mins, sec = hms.split(':')
    return year, month, day, hour, mins, sec, ns


def numpy_datetime_str_to_tup2(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    ts = (numpy_dt - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
    dt = datetime.utcfromtimestamp(ts)

    return dt.year, dt.month, dt.day



import re
from packaging import version
from subprocess import Popen, PIPE


def get_ram_linux():
    vmstat, err = Popen('vmstat -s'.split(), stdout=PIPE, stderr=PIPE).communicate()

    return int(vmstat.splitlines()[0].split()[0])


def get_ram_windows():
    pass


## Version compare
def version_compare(v1, v2) :
    if (v2 is None or v1 is None):
        return None
    r1 = re.search("\\d{4}(\\.\\d+)+", v1)
    r2 = re.search("\\d{4}(\\.\\d+)+", v2)
    if (r2 is None or r1 is None):
        return None
    v1 = version.parse(r1.group(0))
    v2 = version.parse(r2.group(0))
    return -1 if v1 < v2 else 1 if v1 > v2 else 0


def get_array_size(data_size: int, buffer_length: int) -> int:
    """Get the SQream ARRAY size by inner data size and buffer length

    Args:
        data_size: integer with the size of data inside ARRAY, for
          example for INT is 4, for BOOL is 1, etc.
        buffer_length: length of a chunk of buffer connected with one
          array

    Returns:
        An integer representing size of an ARRAY with fixed sized data
    """
    aligned_block_size = (data_size + 1) * 8  # data + 1 byte for null
    div, mod = divmod(buffer_length, aligned_block_size)
    size = div * 8
    if mod:
        size += int((mod - 8) / data_size)
    return size


def false_generator():
    """Generate endless sequence of False values

    Used for providing to zip(data, false_generator()) within data that
    is not nullable, so is_null value will also goes as False
    independent of size of data

    Example:
        >>> for val, is_null in zip([1, 2, 3, 4]], false_generator()):
        ...     print(val,is_null)
        ...
        1 False
        2 False
        3 False
        4 False

    Returns:
        A generator object that produces False value for each iteration
          endlessly
    """
    while True:
        yield False


class Error(Exception):
    pass

class Warning(Exception):
    pass

class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


class ArraysAreDisabled(DatabaseError):
    pass



"""Contain Cursor, which is used to manage the context of a fetch.

Gets statement, send it to SQream then check if statement ready and
send it again for exectuion.
Responsible for both fetching and extracting data.

Should be used only by .connection.Connection
"""

import functools
import json
import logging
import struct

from typing import List, Any, Union

from .globals import BUFFER_SIZE, ROWS_PER_FLUSH, DEFAULT_CHUNKSIZE, \
    FETCH_MANY_DEFAULT, typecodes, type_to_letter, BYTES_PER_FLUSH_LIMIT, TEXT_ITEM_SIZE
from .column_buffer import ColumnBuffer
from .logger import log_and_raise, logger, printdbg
from .utils import NotSupportedError, ProgrammingError, get_array_size, \
    false_generator, ArraysAreDisabled, OperationalError
from .casting import lengths_to_pairs, sq_date_to_py_date, \
    sq_datetime_to_py_datetime, sq_numeric_to_decimal, arr_lengths_to_pairs
from .ping import _start_ping_loop, _end_ping_loop


def _is_null(nullable):
    return nullable == 1


class Cursor:
    """
    Represent a database cursor, which is used to manage the context of
    a fetch operation.

    Cursors created from the same connection are not
    isolated, i.e., any changes done to the database by a cursor are
    immediately visible by the other cursors.

    Base PEP 249 – Python Database API Specification v2.0
    https://peps.python.org/pep-0249/#id12
    """

    def __init__(self, conn, cursors):

        self.conn = conn
        self.cursors = cursors
        self.socket = self.conn.socket
        self.client = self.conn.client
        self.version = self.conn.version
        self.connection_id = self.conn.connection_id
        self.open_statement = False
        self.closed = False
        self.buffer = ColumnBuffer(BUFFER_SIZE)  # flushing buffer every BUFFER_SIZE bytes
        self.stmt_id = None  # For error handling when called out of order
        self.statement_type = None
        self.arraysize = FETCH_MANY_DEFAULT
        self.rowcount = -1  # DB-API property
        self.more_to_fetch = False
        self.ping_loop = self.conn.ping_loop
        self.parsed_rows = []
        self.row_size = 0
        self.rows_per_flush = 0
        self.lastrowid = None
        self.base_connection_closed = False

    def get_statement_type(self):

        return self.statement_type

    def get_statement_id(self):

        return self.stmt_id

    def _execute_sqream_statement(self, stmt):
        self.latest_stmt = stmt

        if self.open_statement:
            self.close_stmt()
        self.open_statement = True

        self.more_to_fetch = True

        self.stmt_id = json.loads(self.client.send_string('{"getStatementId" : "getStatementId"}'))["statementId"]
        stmt_json = json.dumps({"prepareStatement": stmt, "chunkSize": DEFAULT_CHUNKSIZE})
        res = self.client.send_string(stmt_json)

        self.client.validate_response(res, "statementPrepared")
        self.lb_params = json.loads(res)
        if self.lb_params.get('reconnect'):  # Reconnect exists and issued, otherwise False / None

            # Close socket, open a new socket with new port/ip sent be the reconnect response
            self.socket.reconnect(
                self.lb_params['ip'], self.lb_params['port_ssl']
                if self.conn.use_ssl else self.lb_params['port'])

            # Send reconnect and reconstruct messages
            reconnect_str = '{{"service": "{}", "reconnectDatabase":"{}", "connectionId":{}, "listenerId":{},"username":"{}", "password":"{}"}}'.format(
                self.conn.service, self.conn.database, self.conn.connection_id,
                self.lb_params['listener_id'], self.conn.username, self.conn.password)
            self.client.send_string(reconnect_str)
            self.client.send_string('{{"reconstructStatement": {}}}'.format(
                self.stmt_id))

        # Reconnected/reconstructed if needed,  send  execute command
        self.client.validate_response(self.client.send_string('{"execute" : "execute"}'), 'executed')

        # Send queryType message/s
        res = json.loads(self.client.send_string('{"queryTypeIn": "queryTypeIn"}'))
        self.column_list = res.get('queryType', '')

        if not self.column_list:
            res = json.loads(
                self.client.send_string('{"queryTypeOut" : "queryTypeOut"}'))
            self.column_list = res.get('queryTypeNamed', '')
            if not self.column_list:
                self.statement_type = 'DML'
                self.close_stmt()
                return

            self.statement_type = 'SELECT' if self.column_list else 'DML'
            self.result_rows = []
            self.unparsed_row_amount = 0
            self.data_columns = []
        else:
            self.statement_type = 'INSERT'

        # Check if arrays are allowed before executing the rest
        if not self._validate_arrays_usage():
            log_and_raise(
                ArraysAreDisabled, "Arrays are disabled in this connection.")

        # {"isTrueVarChar":false,"nullable":true,"type":["ftInt",4,0]}
        self.col_names, self.col_tvc, self.col_nul, self.col_type_tups = \
            list(zip(*[(col.get("name", ""), col["isTrueVarChar"], col["nullable"], col["type"]) for col in self.column_list]))
        self.col_names_map = {
            name: idx
            for idx, name in enumerate(self.col_names)
        }

        if self.statement_type == 'INSERT':
            # variables should not be defined outside __init__
            # TODO: get rid of definition of attributes outside __init__
            self.col_types = []
            self.col_sizes = []
            self.col_scales = []
            for type_tup in self.col_type_tups:
                is_array = 'ftArray' in type_tup
                offset = 0
                _type = type_tup[0]
                if is_array:
                    # for array other stuff like scale is shifted in type_tup
                    offset = 1
                    # Use tuple for ftArray that will be checked
                    # only in buffer like 'ftArray' in col_type
                    _type = type_tup[0:2]
                self.col_types.append(_type)
                self.col_sizes.append(type_tup[1 + offset] if type_tup[1 + offset] != 0 else TEXT_ITEM_SIZE)
                self.col_scales.append(type_tup[2 + offset])

            self.row_size = sum(self.col_sizes) + sum(
                null for null in self.col_nul if null is True) + sum(tvc for tvc in self.col_tvc if tvc is True)

            self.rows_per_flush = int(ROWS_PER_FLUSH if self.row_size * ROWS_PER_FLUSH <= BYTES_PER_FLUSH_LIMIT else BYTES_PER_FLUSH_LIMIT / self.row_size)
            self.buffer.clear()

        if self.statement_type == 'SELECT':
            self.parsed_rows = []
            self.parsed_row_amount = 0

        if logger.isEnabledFor(logging.INFO):
            logger.info \
                (f'Executing statement over connection {self.conn.connection_id} with statement id {self.stmt_id}:\n{stmt}')

    def _validate_arrays_usage(self) -> bool:
        """
        Checks if the executing statement uses arrays and if they are allowed.

        Returns:
            bool: False if arrays are not allowed by connection, but used in
              statement, True otherwise.
        """
        if not self.conn.allow_array:
            for col in self.column_list:
                if "ftArray" in col["type"]:
                    return False
        return True

    def _fill_description(self):
        """Getting parameters for the cursor's 'description' attribute, even for
           a query that returns no rows. For each column, this includes:
           (name, type_code, display_size, internal_size, precision, scale, null_ok)"""

        if self.statement_type != 'SELECT':
            self.description = None
            return self.description

        self.description = []
        for col_name, col_nullalbe, col_type_tup in zip(
                self.col_names, self.col_nul, self.col_type_tups):
            type_code = typecodes[
                col_type_tup[0]]  # Convert SQream type to DBAPI identifier
            display_size = internal_size = col_type_tup[
                1]  # Check if other size is available from API
            precision = 38
            scale = col_type_tup[2]

            self.description.append(
                (col_name, type_code, display_size, internal_size, precision,
                 scale, col_nullalbe))

        return self.description

    def _send_columns(self, cols=None, capacity=None):
        """Perform network insert - "put" json, header, binarized columns. Used by executemany()"""

        cols = cols or self.cols
        cols = cols if isinstance(cols, (list, tuple, set, dict)) else list(cols)

        capacity = capacity or self.capacity
        # Send columns and metadata to be packed into our buffer
        packed_cols = self.buffer.pack_columns(cols, capacity, self.col_types,
                                               self.col_sizes, self.col_nul,
                                               self.col_tvc, self.col_scales)
        byte_count = functools.reduce(lambda c, n: c + len(n), packed_cols, 0)

        # Stop and start ping is must between sending message to the server, this is part of the protocol.
        _end_ping_loop(self.ping_loop)
        # Sending put message and binary header
        self.client.send_string(f'{{"put":{capacity}}}', False)
        self.ping_loop = _start_ping_loop(self.client, self.socket)

        _end_ping_loop(self.ping_loop)
        self.socket.send((self.client.generate_message_header(byte_count, False)))
        self.ping_loop = _start_ping_loop(self.client, self.socket)

        # Sending packed data (binary buffer)
        _end_ping_loop(self.ping_loop)
        for packed_col in packed_cols:
            printdbg("Packed data sent:", packed_col)
            self.socket.send((packed_col))

        self.client.validate_response(self.client.get_response(), '{"putted":"putted"}')
        self.ping_loop = _start_ping_loop(self.client, self.socket)

    def _fetch(self):

        # JSON correspondence
        res = self.client.send_string('{"fetch" : "fetch"}')
        self.client.validate_response(res, "colSzs")
        fetch_meta = json.loads(res)
        num_rows_fetched, column_sizes = fetch_meta['rows'], fetch_meta['colSzs']
        if num_rows_fetched == 0:
            self.close_stmt()
            return num_rows_fetched

        # Get preceding header
        self.client.receive(10)

        # Get data as memoryviews of bytearrays.
        unsorted_data_columns = [memoryview(self.client.receive(size)) for idx, size in enumerate(column_sizes)]

        # Sort by columns, taking a memoryview and casting to the proper type
        self.data_columns = []

        for type_tup, nullable, tvc in zip(self.col_type_tups, self.col_nul,
                                           self.col_tvc):
            column = {'nullable': False, 'true_nvarchar': False}

            is_array = 'ftArray' in type_tup

            if nullable:
                column['nullable'] = unsorted_data_columns.pop(0)
            if is_array:
                column['array_lengths'] = unsorted_data_columns\
                    .pop(0).cast('i')
            elif tvc:
                column['true_nvarchar'] = unsorted_data_columns\
                    .pop(0).cast('i')

            column['data_column'] = unsorted_data_columns.pop(0)

            if type_tup[0] in ('ftVarchar', 'ftBlob', 'ftNumeric'):
                column['data_column'] = column['data_column'].tobytes()
            elif not is_array:
                column['data_column'] = column['data_column'].cast(
                    type_to_letter[type_tup[0]])

            self.data_columns.append(column)

        self.unparsed_row_amount = num_rows_fetched

        return num_rows_fetched

    def _parse_fetched_cols(self):
        """Used by _fetch_and_parse"""

        self.extracted_cols = []

        if not self.data_columns:
            return self.extracted_cols

        for idx, raw_col_data in enumerate(self.data_columns):

            # Extract data according to column type
            if self.col_type_tups[idx][0] == "ftArray":
                col = self._extract_array(idx, raw_col_data)

            elif self.col_tvc[idx]:  # nvarchar
                col = self._extract_nvarchar(idx, raw_col_data)

            elif self.col_type_tups[idx][0] == "ftVarchar":
                col = self._extract_varchar(idx, raw_col_data)

            elif self.col_type_tups[idx][0] == "ftDate":
                col = self._extract_date(idx, raw_col_data)

            elif self.col_type_tups[idx][0] == "ftDateTime":
                col = self._extract_datetime(idx, raw_col_data)

            elif self.col_type_tups[idx][0] == "ftNumeric":
                col = self._extract_numeric(idx, raw_col_data)
            else:
                col = self._extract_datatype(idx, raw_col_data)

            self.extracted_cols.append(col)

        # Done with the raw data buffers
        self.unparsed_row_amount = 0
        self.data_columns = []

        return self.extracted_cols

    def _fetch_and_parse(self, requested_row_amount, data_as='rows'):
        ''' See if this amount of data is available or a fetch from sqream is required
            -1 - fetch all available data. Used by fetchmany() '''

        if data_as == 'rows':
            while (requested_row_amount > len(self.parsed_rows)
                   or requested_row_amount == -1) and self.more_to_fetch:
                self.more_to_fetch = bool(self._fetch())  # _fetch() updates self.unparsed_row_amount

                self.parsed_rows.extend(zip(*self._parse_fetched_cols()))

    def execute(self, query, params=None):
        """Execute a statement. Parameters are not supported"""

        if self.base_connection_closed:
            self.conn._verify_con_open()
        else:
            self.conn._verify_cur_open()
        if params:

            log_and_raise(ProgrammingError, "Parametered queries not supported. \
                If this is an insert query, use executemany() with the data rows as the parameter")

        else:
            self._execute_sqream_statement(query)

        self._fill_description()
        self.rows_fetched = 0
        self.rows_returned = 0
        return self

    def executemany(self, query, rows_or_cols=None, data_as='rows', amount=None):
        ''' Execute a statement, including parametered data insert '''

        # self._verify_open()
        self.execute(query)

        if rows_or_cols is None:
            return self

        if data_as =='alchemy_flat_list':
            # Unflatten SQLalchemy data list
            row_len = len(self.column_list)
            rows_or_cols = [rows_or_cols[i: i +row_len] for i in range(0, len(rows_or_cols), row_len)]
            data_as = 'rows'

        if 'numpy' in repr(type(rows_or_cols[0])):
            data_as = 'numpy'

        # Network insert starts here if data was passed
        column_lengths = [len(row_or_col) for row_or_col in rows_or_cols]
        if column_lengths.count(column_lengths[0]) != len(column_lengths):
            log_and_raise(ProgrammingError,
                          "Incosistent data sequences passed for inserting. Please use rows/columns of consistent length"
                          )
        if data_as == 'rows':
            self.capacity = amount or len(rows_or_cols)
            self.cols = list(zip(*rows_or_cols))
        else:
            self.cols = rows_or_cols
            self.capacity = len(self.cols)

        # Slice a chunk of columns and pass to _send_columns()
        start_idx = 0
        while self.cols != [()]:
            col_chunk = [col[start_idx:start_idx + self.rows_per_flush] for col in self.cols]
            chunk_len = len(col_chunk[0])
            if chunk_len == 0:
                break
            self._send_columns(col_chunk, chunk_len)
            start_idx += self.rows_per_flush
            del col_chunk

            if logger.isEnabledFor(logging.INFO):
                logger.info(f'Sent {chunk_len} rows of data')

        self.close_stmt()

        return self

    def fetchmany(self, size=None, data_as='rows', fetchone=False):
        ''' Fetch an amount of result rows '''

        size = size or self.arraysize
        # self._verify_open()
        if self.statement_type not in (None, 'SELECT'):
            log_and_raise(ProgrammingError ,'No open statement while attempting fetch operation')

        if self.more_to_fetch is False:
            # All data from server for this select statement was fetched
            if len(self.parsed_rows) == 0:
                # Nothing
                return [] if not fetchone else None
        else:
            self._fetch_and_parse(size, data_as)

        # Get relevant part of parsed rows and reduce storage and counter
        if data_as == 'rows':
            res = self.parsed_rows[0:size if size != -1 else None]
            del self.parsed_rows[:size if size != -1 else None]

        if logger.isEnabledFor(logging.INFO):
            logger.info(f'Fetched {size} rows')

        return (res if res else []) if not fetchone else (res[0] if res else None)

    def fetchone(self, data_as='rows'):
        """Fetch one result row"""

        if data_as not in ('rows',):
            log_and_raise(ProgrammingError, "Bad argument to fetchone()")

        return self.fetchmany(1, data_as, fetchone=True)

    def fetchall(self, data_as='rows'):
        """Fetch all result rows"""

        if data_as not in ('rows',):
            log_and_raise(ProgrammingError, "Bad argument to fetchall()")

        return self.fetchmany(-1, data_as)

    # DB-API Do nothing (for now) methods
    # -----------------------------------

    def nextset(self):
        """No multiple result sets so currently always returns None"""
        log_and_raise(NotSupportedError, "Nextset is not supported")

    def setinputsizes(self, sizes):
        log_and_raise(NotSupportedError, "Setinputsizes is not supported")

    def setoutputsize(self, size, column=None):
        log_and_raise(NotSupportedError, "Setoutputsize is not supported")

    # def csv_to_table(self, csv_path, table_name, read=None, parse=None, convert=None, con=None, auto_infer=False,
    #                  delimiter="|"):
    #     ' Pyarrow CSV reader documentation: https://arrow.apache.org/docs/python/generated/pyarrow.csv.read_csv.html '
    #
    #     if not ARROW:
    #         return "Optional pyarrow dependency not found. To install: pip3 install pyarrow"
    #
    #     sqream_to_pa = {
    #         'ftBool':     pa.bool_(),
    #         'ftUByte':    pa.uint8(),
    #         'ftShort':    pa.int16(),
    #         'ftInt':      pa.int32(),
    #         'ftLong':     pa.int64(),
    #         'ftFloat':    pa.float32(),
    #         'ftDouble':   pa.float64(),
    #         'ftDate':     pa.timestamp('ns'),
    #         'ftDateTime': pa.timestamp('ns'),
    #         'ftVarchar':  pa.string(),
    #         'ftBlob':     pa.utf8(),
    #         'ftNumeric':  pa.decimal128(28, 11)
    #     }
    #
    #     start = time.time()
    #     # Get table metadata
    #     self.execute(f'select * from {table_name} where 1=0')
    #
    #     # Map column names to pyarrow types and set Arrow's CSV parameters
    #     sqream_col_types = [col_type[0] for col_type in self.col_type_tups]
    #     column_types = zip(self.col_names, [sqream_to_pa[col_type[0]] for col_type in self.col_type_tups])
    #     read = read or csv.ReadOptions(column_names=self.col_names)
    #     parse = parse or csv.ParseOptions(delimiter=delimiter)
    #     convert = convert or csv.ConvertOptions(column_types=None if auto_infer else column_types,
    #                                             null_values=["\\N"])
    #
    #     # Read CSV to in-memory arrow format
    #     csv_arrow = csv.read_csv(csv_path, read_options=read, parse_options=parse, convert_options=convert).\
    #         combine_chunks()
    #     num_chunks = len(csv_arrow[0].chunks)
    #     numpy_cols = []
    #
    #     # For each column, get the numpy representation for quick packing
    #     for col_type, col in zip(sqream_col_types, csv_arrow):
    #         # Only one chunk after combine_chunks()
    #         col = col.chunks[0]
    #         if col_type in  ('ftVarchar', 'ftBlob', 'ftDate', 'ftDateTime', 'ftNumeric'):
    #             col = col.to_pandas()
    #         else:
    #             col = col.to_numpy()
    #
    #         numpy_cols.append(col)
    #
    #     print(f'total loading csv: {time.time( ) -start}')
    #     start = time.time()
    #
    #     # Insert columns into SQream
    #     col_num = csv_arrow.shape[1]
    #     self.executemany(f'insert into {table_name} values ({"?, " *(col_num-1)}?)', numpy_cols)
    #     print(f'total inserting csv: {time.time( ) -start}')

    ## Closing

    def close_stmt(self) -> None:
        """Closes open statement with SQREAM

        Raises:
            ProgrammingError: If server responds with invalid JSON
            ProgrammingError: If server responds with valid JSON,
              but it is not object
            OperationalError: If server responds with "error" key in JSON
        """
        if self.open_statement:
            raw = self.client.send_string(
                '{"closeStatement": "closeStatement"}')

            # Check errors in response of the server
            if raw:
                try:
                    response = json.loads(raw)
                except json.decoder.JSONDecodeError:
                    log_and_raise(
                        ProgrammingError,
                        f"Could not parse server response: {raw}"
                    )
                if not isinstance(response, dict):
                    log_and_raise(
                        ProgrammingError,
                        f"Unexpected server response: {raw}"
                    )
                if "error" in response:
                    log_and_raise(OperationalError, response["error"])

            self.open_statement = False

            if logger.isEnabledFor(logging.INFO):
                logger.info(f'Done executing statement {self.stmt_id} over connection {self.conn.connection_id}')

    def _extract_nvarchar(self, idx, raw_col_data):
        if self.col_nul[idx]:
            col = [None if (_is_null(n)) else raw_col_data['data_column'][start:end].decode('utf8') for (start, end), n
                   in
                   zip(lengths_to_pairs(raw_col_data['true_nvarchar']), raw_col_data['nullable'])]
        else:
            col = [
                raw_col_data['data_column'][start:end].decode('utf8')
                for (start, end) in lengths_to_pairs(raw_col_data['true_nvarchar'])
            ]
        return col

    def _extract_varchar(self, idx, raw_col_data):
        varchar_size = self.col_type_tups[idx][1]
        if self.col_nul[idx]:
            col = []
            offset = 0
            for idx in raw_col_data['nullable']:
                if _is_null(idx):
                    col.append(None)
                    offset = offset + varchar_size
                else:
                    col.append(raw_col_data['data_column'][offset:offset + varchar_size].decode(self.conn.varchar_enc,
                                                                                                "ignore").replace(
                        '\x00', '').rstrip())
                    offset = offset + varchar_size
        else:
            col = [
                raw_col_data['data_column'][idx:idx + varchar_size].decode(
                    self.conn.varchar_enc, "ignore").replace('\x00', '').rstrip()
                for idx in range(0, len(raw_col_data['data_column']), varchar_size)
            ]
        return col

    def _extract_date(self, idx, raw_col_data):
        if self.col_nul[idx]:
            col = [sq_date_to_py_date(d, is_null=_is_null(n)) for d, n in
                   zip(raw_col_data['data_column'], raw_col_data['nullable'])]
        else:
            col = [sq_date_to_py_date(d) for d in raw_col_data['data_column']]
        return col

    def _extract_datetime(self, idx, raw_col_data):
        if self.col_nul[idx]:
            col = [sq_datetime_to_py_datetime(d, is_null=_is_null(n)) for d, n in
                   zip(raw_col_data['data_column'], raw_col_data['nullable'])]
        else:
            col = [sq_datetime_to_py_datetime(d) for d in raw_col_data['data_column']]
        return col

    def _extract_numeric(self, idx, raw_col_data):
        scale = self.col_type_tups[idx][2]
        if self.col_nul[idx]:
            col = [
                sq_numeric_to_decimal(raw_col_data['data_column'][idx:idx + 16], scale, is_null=_is_null(n))
                for idx, n in zip(range(0, len(raw_col_data['data_column']), 16), raw_col_data['nullable'])
            ]
        else:
            col = [
                sq_numeric_to_decimal(raw_col_data['data_column'][idx:idx + 16], scale)
                for idx in range(0, len(raw_col_data['data_column']), 16)
            ]
        return col

    def _extract_datatype(self, idx, raw_col_data):
        if self.col_nul[idx]:
            col = [None if _is_null(n) else d for d, n in zip(raw_col_data['data_column'], raw_col_data['nullable'])]
        else:
            col = raw_col_data['data_column']
        return col

    def _extract_array(
            self, idx: int, raw_col_data: memoryview) -> List[List[Any]]:
        """Extract array data from buffer

        Args:
            idx: integer index of extracting column in response
            raw_col_data: memoryview (bytes represenation) of data of
              column

        Returns:
            A list with Arrays of data. Array represented as python
            lists also.

            [[1, 5, 7], None, [31, 2, None, 6]]
        """
        sub_type_tup = self.col_type_tups[idx]
        typecode = typecodes.get(sub_type_tup[1])

        if typecode == "STRING":
            return self._extract_unfixed_array(raw_col_data)

        if typecode not in ("NUMBER", "DATETIME"):
            log_and_raise(
                NotSupportedError,
                f'Array of "{sub_type_tup[1]}" is not supported',
            )

        return self._extract_fixed_array(idx, raw_col_data)

    def _extract_fixed_array(
            self, idx: int, raw_col_data: memoryview) -> List[List[Any]]:
        """Extract array with data of fixed size

        Extract array from binary data of an Array with types of fixed
        size (BOOL, TINYINT, SMALLINT, INT, BIGINT, REAL, DOUBLE,
        NUMERIC, DATE, DATETIME). But not with TEXT

        Raw data contains binary data of nulls at each index in array
        and data separated by optional padding (trailing zeros at the
        end for portions of data whose lengths are not dividable by 8)

        Example for binary data for 1 row of boolean array[true, null,
        false]:
        `010 00000 100` -> replace paddings with _ `010_____100` where
        `010` are flag of null data inside array. Then `00000` is a
        padding to make lengths of data about nulls to be dividable by 8
        in case of array of length 8, 16, 24, 32 ... there won't be a
        padding then `100` is a binary representation of 3 boolean
        values itself

        Args:
            idx: integer index of extracting column in response
            raw_col_data: memoryview (bytes represenation) of data of
              column

        Returns:
            A list with Arrays of data. Array represented as python
            lists also.

            [[1, 5, 7], None, [31, 2, None, 6]]
        """
        buffer = raw_col_data['data_column']
        nulls_buffer = raw_col_data['nullable'] or false_generator()

        # Calculate size based on data_format
        data_size = struct.calcsize(type_to_letter[self.col_type_tups[idx][1]])
        trasform = self._get_trasform_func(idx)

        def _get_array(data: memoryview, nulls: memoryview, arr_size: int):
            """Construct one single array from data of type with fixed size"""
            # Do not use direct data.cast(data_format) to skip nulls
            # and process all data (including Numeric) by the same pattern
            return [
                trasform(data[i * data_size:(i + 1) * data_size], nulls[i])
                for i in range(arr_size)
            ]

        col = []
        start = 0
        for buf_len, null in zip(raw_col_data['array_lengths'], nulls_buffer):
            if null:
                col.append(None)
            else:
                array_size = get_array_size(data_size, buf_len)
                padding = (8 - array_size % 8) % 8

                # Slices of memoryview do not copy underlying data
                data = buffer[start + array_size + padding:start + buf_len]
                nulls = buffer[start:start + array_size]

                col.append(_get_array(data, nulls, array_size))
            start += buf_len
        return col

    def _extract_unfixed_array(
            self, raw_col_data: memoryview) -> List[List[Union[str, None]]]:
        """Extract array with data of unfixed size

        Extract array from binary data of an Array with types of TEXT
        - unfixed size

        Contains 8 bytes (long) that contains length of whole array
        (including nulls), binary data of nulls at each index in array
        and data separated by optional padding. Data here represents
        chunked info of each element inside array

        At the beginning, the data contains **cumulative** lengths
        (however is better to say indexes of their ends at data buffer)
        of all data strings (+ their paddings) of array as integers.
        The number of those int lengths is equal to the array length
        (those was in 8 bytes above) and because int take 4 bytes it all
        takes N * 4 bytes. Then if it is not divisible by 8 -> + padding
        Then the strings data also separated by optional padding

        Example for binary data for 1 row of text array['ABC','ABCDEF',null]:
        (padding zeros replaced with _)
        Whole buffer data: `3000000 001_____ 3000 14000 16000 ____ `
                           `65 66 67 _____ 65 66 67 68 69 70 __`
        Length of array: `3000000` -> long 3
        Nulls: `001_____`
        Length of strings: `3000 14000 16000 ____` -> 3,14,16 + padding
        Strings: `65, 66, 67, _____ 65, 66, 67, 68, 69, 70, __`
        L1 = 3, so [0 - 3) is string `65 66 67` -> ABC, padding P1=5
        L2 = 14 (which is L1 + padding + current_length), so
        current_length = L2 - (L1 + P1) = 14 - (5 + 3) = 6, P2=2
        => [5 + 3, 14) is string `65, 66, 67, 68, 69, 70` -> ABCDEF
        L3 = 16 => current_length = L3 - (L2 + P2) = 16 - (14 + 2) = 0
        thus string is empty, and considering Nulls -> it is a null

        Args:
            idx: integer index of extracting column in response
            raw_col_data: memoryview (bytes represenation) of data of
              column

        Returns:
            A list with Arrays of data. Array represented as python
            lists also.

            [["ABC", "ABCDEF", None], None, ["A", None, ""]]
        """
        # Code duplication with _extract_fixed_array could be eliminated
        # using template method with using separate Extractor classes
        buffer = raw_col_data['data_column']
        nulls_buffer = raw_col_data['nullable'] or false_generator()

        def _get_array(data: memoryview, nulls: memoryview, dlen: memoryview):
            """Construct one single array from data with dlen right bounds"""
            arr = []
            # lengths_to_pairs is not appropriate due to differences
            # in lengths representation
            for is_null, (strt, end) in zip(nulls, arr_lengths_to_pairs(dlen)):
                if is_null:
                    arr.append(None)
                else:
                    arr.append(data[strt:end].tobytes().decode('utf8'))
            return arr

        col = []
        start = 0
        for buf_len, null in zip(raw_col_data['array_lengths'], nulls_buffer):
            if null:
                col.append(None)
            elif not buf_len:
                col.append([])
            else:
                array_size = buffer[start: start + 8].cast('q')[0]  # Long
                padding = (8 - array_size % 8) % 8
                cur = start + 8 + array_size + padding
                # data lengths
                d_len = buffer[cur:cur + array_size * 4].cast('i')
                cur += (array_size + array_size % 2) * 4

                # Slices of memoryview do not copy underlying data
                data = buffer[cur:start + buf_len]
                nulls = buffer[start + 8:start + 8 + array_size]

                col.append(_get_array(data, nulls, d_len))

            start += buf_len
        return col

    def _get_trasform_func(self, idx: int) -> callable:
        """Provide function for casting bytes data to real data

        Args:
            idx: integer index of extracting column in response

        Returns:
            A function that cast simple portion of data to appropriate
            value.
        """
        type_tup = self.col_type_tups[idx]

        # Array's type_tup differs from others by adding extra string
        # at the beginning
        offset = 1 if type_tup[0] == 'ftArray' else 0
        data_format = type_to_letter[type_tup[offset]]
        wrappers = {
            'ftDate': sq_date_to_py_date,
            'ftDateTime': sq_datetime_to_py_datetime
        }

        if type_tup[offset] == 'ftNumeric':
            scale = type_tup[offset + 2]

            def cast(data):
                return sq_numeric_to_decimal(data, scale)
        elif type_tup[offset] in wrappers:
            def cast(data):
                return wrappers[type_tup[offset]](data.cast(data_format)[0])
        else:
            def cast(data):
                return data.cast(data_format)[0]

        def trasform(mem: memoryview, is_null: bool = False):
            return None if is_null else cast(mem)

        return trasform

    def close(self):
        self.close_stmt()
        self.conn.cur_closed = True
        self.conn.close_connection()
        self.closed = True
        self.buffer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def __iter__(self):
        for item in self.fetchall():
            yield item



from setuptools import setup, find_packages

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup_params = dict(
    name =          'pysqream',
    version =       '5.1.0',
    description =   'DB-API connector for SQream DB', 
    long_description = long_description,
    url = "https://github.com/SQream/pysqream",
    author = "SQream",
    author_email = "info@sqream.com",
    packages = find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    keywords = 'database db-api sqream sqreamdb',
    python_requires = '>=3.9',
    install_requires=["numpy==1.26.4", "packaging>=23.0", "pyarrow==15.0.0",
                      "pandas==2.2.2", "Faker>=26.0.0", "numexpr==2.8.4"]
)

if __name__ == '__main__':
    setup(**setup_params)
